@@ -46,6 +46,7 @@
 #include "nsIDOMEventListener.h"
 #include "nsIDOMEventTarget.h"
 #include "nsIDOMHTMLAnchorElement.h"
+#include "nsIDOMHTMLImageElement.h"
 #include "nsIDOMHTMLInputElement.h"
 #include "nsIDOMHTMLSelectElement.h"
 #include "nsIDOMNSEvent.h"
@@ -53,6 +54,7 @@
 #include "nsIDOMXULMenuListElement.h"
 #include "nsIDOMXULMultSelectCntrlEl.h"
 #include "nsIDOMXULSelectCntrlItemEl.h"
+#include "nsIDOMXULPopupElement.h"
 #include "nsIDocument.h"
 #include "nsIEventListenerManager.h"
 #include "nsIHTMLDocument.h"
@@ -238,6 +240,10 @@ nsresult nsRootAccessible::AddEventListeners()
     rv = target->AddEventListener(NS_LITERAL_STRING("select"), NS_STATIC_CAST(nsIDOMFormListener*, this), PR_TRUE);
     NS_ENSURE_SUCCESS(rv, rv);
 
+    // capture NameChange events (fired whenever name changes, immediately after, whether focus moves or not)
+    rv = target->AddEventListener(NS_LITERAL_STRING("NameChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
+    NS_ENSURE_SUCCESS(rv, rv);
+
     // capture ValueChange events (fired whenever value changes, immediately after, whether focus moves or not)
     rv = target->AddEventListener(NS_LITERAL_STRING("ValueChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     NS_ENSURE_SUCCESS(rv, rv);
@@ -246,6 +252,10 @@ nsresult nsRootAccessible::AddEventListeners()
     rv = target->AddEventListener(NS_LITERAL_STRING("AlertActive"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     NS_ENSURE_SUCCESS(rv, rv);
 
+    // add ourself as a TreeViewChanged listener (custom event fired in tree.xml)
+    rv = target->AddEventListener(NS_LITERAL_STRING("TreeViewChanged"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
+    NS_ENSURE_SUCCESS(rv, rv);
+
     // add ourself as a OpenStateChange listener (custom event fired in tree.xml)
     rv = target->AddEventListener(NS_LITERAL_STRING("OpenStateChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     NS_ENSURE_SUCCESS(rv, rv);
@@ -287,10 +297,6 @@ nsresult nsRootAccessible::AddEventListeners()
                                            NS_STATIC_CAST(nsIDOMXULListener*, this), 
                                            PR_TRUE);
     NS_ENSURE_SUCCESS(rv, rv);
-    target->AddEventListener(NS_LITERAL_STRING("pageshow"), 
-                             NS_STATIC_CAST(nsIDOMXULListener*, this), 
-                             PR_TRUE);
-    NS_ENSURE_SUCCESS(rv, rv);
   }
 
   if (!mCaretAccessible)
@@ -313,8 +319,10 @@ nsresult nsRootAccessible::RemoveEventListeners()
   if (target) { 
     target->RemoveEventListener(NS_LITERAL_STRING("focus"), NS_STATIC_CAST(nsIDOMFocusListener*, this), PR_TRUE);
     target->RemoveEventListener(NS_LITERAL_STRING("select"), NS_STATIC_CAST(nsIDOMFormListener*, this), PR_TRUE);
+    target->RemoveEventListener(NS_LITERAL_STRING("NameChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     target->RemoveEventListener(NS_LITERAL_STRING("ValueChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     target->RemoveEventListener(NS_LITERAL_STRING("AlertActive"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
+    target->RemoveEventListener(NS_LITERAL_STRING("TreeViewChanged"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     target->RemoveEventListener(NS_LITERAL_STRING("OpenStateChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     target->RemoveEventListener(NS_LITERAL_STRING("CheckboxStateChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
     target->RemoveEventListener(NS_LITERAL_STRING("RadioStateChange"), NS_STATIC_CAST(nsIDOMXULListener*, this), PR_TRUE);
@@ -332,9 +340,6 @@ nsresult nsRootAccessible::RemoveEventListeners()
     target->RemoveEventListener(NS_LITERAL_STRING("pagehide"), 
                                 NS_STATIC_CAST(nsIDOMXULListener*, this), 
                                 PR_TRUE);
-    target->RemoveEventListener(NS_LITERAL_STRING("pageshow"), 
-                                NS_STATIC_CAST(nsIDOMXULListener*, this), 
-                                PR_TRUE);
   }
 
   if (mCaretAccessible) {
@@ -421,14 +426,30 @@ void nsRootAccessible::TryFireEarlyLoadEvent(nsIAccessible *aAccessible, nsIDOMN
 
 void nsRootAccessible::FireAccessibleFocusEvent(nsIAccessible *aAccessible,
                                                 nsIDOMNode *aNode,
+                                                nsIDOMEvent *aFocusEvent,
                                                 PRBool aForceEvent)
 {
   NS_ASSERTION(aAccessible, "Attempted to fire focus event for no accessible");
 
-  // Fire focus only if it changes, but always fire focus events for menu items
-  // Also always fire for XUL tree items, because they always use the same node for
-  // the tree container itself
-  // XXX use optional aForceIt bool param
+  if (mCaretAccessible) {
+    nsCOMPtr<nsIDOMNSEvent> nsevent(do_QueryInterface(aFocusEvent));
+    if (nsevent) {
+      // Use the originally focused node where the selection lives.
+      // For example, use the anonymous HTML:input instead of the containing
+      // XUL:textbox. In this case, sometimes it is a later focus event
+      // which points to the actual anonymous child with focus, so to be safe 
+      // we need to reset the selection listener every time.
+      // This happens because when some bindings handle focus, they retarget
+      // focus to the appropriate child inside of themselves, but DOM focus
+      // stays outside on that binding parent.
+      nsCOMPtr<nsIDOMEventTarget> domEventTarget;
+      nsevent->GetOriginalTarget(getter_AddRefs(domEventTarget));
+      nsCOMPtr<nsIDOMNode> realFocusedNode = do_QueryInterface(domEventTarget);
+      mCaretAccessible->AttachNewSelectionListener(realFocusedNode);
+    }
+  }
+
+  // Fire focus only if it changes, but always fire focus events when aForceEvent == PR_TRUE
   if (gLastFocusedNode == aNode && !aForceEvent) {
     return;
   }
@@ -462,9 +483,6 @@ void nsRootAccessible::FireAccessibleFocusEvent(nsIAccessible *aAccessible,
 
   privateAccessible->FireToolkitEvent(nsIAccessibleEvent::EVENT_FOCUS,
                                       aAccessible, nsnull);
-
-  if (mCaretAccessible)
-    mCaretAccessible->AttachNewSelectionListener(aNode);
 }
 
 void nsRootAccessible::FireCurrentFocusEvent()
@@ -530,11 +548,11 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
   targetNode->GetLocalName(localName);
 #ifdef DEBUG_aleventhal
   // Very useful for debugging, please leave this here.
-  if (eventType.LowerCaseEqualsLiteral("dommenuitemactive")) {
+  if (eventType.LowerCaseEqualsLiteral("popupshown")) {
     printf("\ndebugging dommenuitemactive events for %s", NS_ConvertUCS2toUTF8(localName).get());
   }
-  if (localName.EqualsIgnoreCase("tree")) {
-    printf("\ndebugging events in tree, event is %s", NS_ConvertUCS2toUTF8(eventType).get());
+  if (localName.EqualsIgnoreCase("popup")) {
+    printf("\ndebugging events in popup, event is %s", NS_ConvertUCS2toUTF8(eventType).get());
   }
   if (localName.EqualsIgnoreCase("select")) {
     printf("\ndebugging events in select, event is %s", NS_ConvertUCS2toUTF8(eventType).get());
@@ -545,6 +563,25 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
 
 #ifdef MOZ_ACCESSIBILITY_ATK
   nsCOMPtr<nsIDOMHTMLAnchorElement> anchorElement(do_QueryInterface(targetNode));
+  // For ATK, check whether this link is for an image.
+  // If so, fire event for the image.
+  if (anchorElement) {
+    nsCOMPtr<nsIDOMNode> childNode;
+    if (NS_SUCCEEDED(targetNode->GetFirstChild(getter_AddRefs(childNode)))) {
+      while (childNode) {
+        nsCOMPtr<nsIDOMHTMLImageElement> imgNode(do_QueryInterface(childNode));
+        if (imgNode) {
+          anchorElement = nsnull; // ignore the link
+          targetNode = childNode; // only fire event for image
+          break;
+        }
+        nsCOMPtr<nsIDOMNode> tmpNode;
+        tmpNode.swap(childNode);
+        tmpNode->GetNextSibling(getter_AddRefs(childNode));
+      }
+    }
+  }
+
   if (anchorElement) {
     nsCOMPtr<nsIDOMNode> blockNode;
     // For ATK, we don't create any individual object for hyperlink, use its parent who has block frame instead
@@ -557,7 +594,7 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     return NS_OK;
   }
       
-  if (eventType.LowerCaseEqualsLiteral("unload")) {
+  if (eventType.LowerCaseEqualsLiteral("pagehide")) {
     // Only get cached accessible for pagehide -- so that we don't create it
     // just to destroy it.
     nsCOMPtr<nsIWeakReference> weakShell(do_GetWeakReference(eventShell));
@@ -570,6 +607,24 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     return NS_OK;
   }
 
+  if (eventType.LowerCaseEqualsLiteral("popupshown")) {
+    // Fire menupopupstart events after a delay so that ancestor views
+    // are visible, otherwise an accessible cannot be created for the
+    // popup and the accessibility toolkit event can't be fired.
+    nsCOMPtr<nsIDOMXULPopupElement> popup(do_QueryInterface(targetNode));
+    if (popup) {
+      return FireDelayedToolkitEvent(nsIAccessibleEvent::EVENT_MENUPOPUPSTART,
+                                    targetNode, nsnull);
+    }
+  }
+
+  if (eventType.EqualsLiteral("TreeViewChanged")) {
+    NS_ENSURE_TRUE(localName.EqualsLiteral("tree"), NS_OK);
+    nsCOMPtr<nsIContent> treeContent = do_QueryInterface(targetNode);
+    return mAccService->InvalidateSubtreeFor(eventShell, treeContent,
+                                             nsIAccessibleEvent::EVENT_REORDER);
+  }
+
   nsCOMPtr<nsIAccessible> accessible;
   if (NS_FAILED(mAccService->GetAccessibleInShell(targetNode, eventShell,
                                                   getter_AddRefs(accessible))))
@@ -613,10 +668,14 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     }
   else if (treeItemAccessible) {
     if (eventType.LowerCaseEqualsLiteral("focus")) {
-      FireAccessibleFocusEvent(accessible, targetNode); // Tree has focus
+      FireAccessibleFocusEvent(accessible, targetNode, aEvent); // Tree has focus
     }
     else if (eventType.LowerCaseEqualsLiteral("dommenuitemactive")) {
-      FireAccessibleFocusEvent(treeItemAccessible, targetNode, PR_TRUE);
+      FireAccessibleFocusEvent(treeItemAccessible, targetNode, aEvent, PR_TRUE);
+    }
+    else if (eventType.LowerCaseEqualsLiteral("namechange")) {
+      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_NAME_CHANGE, 
+                                accessible, nsnull);
     }
     else if (eventType.LowerCaseEqualsLiteral("select")) {
       // If multiselect tree, we should fire selectionadd or selection removed
@@ -659,7 +718,7 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
 
     // Only fire focus event for DOMMenuItemActive is not inside collapsed popup
     if (0 == (containerState & STATE_COLLAPSED)) {
-      FireAccessibleFocusEvent(accessible, targetNode, PR_TRUE);
+      FireAccessibleFocusEvent(accessible, targetNode, aEvent, PR_TRUE);
     }
   }
   else if (eventType.LowerCaseEqualsLiteral("focus")) {
@@ -688,9 +747,16 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     if (accessible == this) {
       // Top level window focus events already automatically fired by MSAA
       // based on HWND activities. Don't fire the extra focus event.
+      NS_IF_RELEASE(gLastFocusedNode);
+      gLastFocusedNode = mDOMNode;
+      NS_IF_ADDREF(gLastFocusedNode);
       return NS_OK;
     }
-    FireAccessibleFocusEvent(accessible, targetNode);
+    FireAccessibleFocusEvent(accessible, targetNode, aEvent);
+  }
+  else if (eventType.LowerCaseEqualsLiteral("namechange")) {
+    privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_NAME_CHANGE, 
+                              accessible, nsnull);
   }
   else if (eventType.EqualsLiteral("ValueChange")) {
     PRUint32 role;
@@ -712,17 +778,12 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
                               accessible, nsnull);
   }
   else if (eventType.LowerCaseEqualsLiteral("radiostatechange") ) {
-    // first the XUL radio buttons
-    if (targetNode &&
-        NS_SUCCEEDED(mAccService->GetAccessibleInShell(targetNode, eventShell,
-                                                       getter_AddRefs(accessible)))) {
-      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_STATE_CHANGE, 
-                                accessible, nsnull);
-      FireAccessibleFocusEvent(accessible, targetNode);
-    }
-    else { // for the html radio buttons -- apparently the focus code just works. :-)
-      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_STATE_CHANGE, 
-                                accessible, nsnull);
+    privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_STATE_CHANGE, 
+                              accessible, nsnull);
+    PRUint32 finalState;
+    accessible->GetFinalState(&finalState);
+    if (finalState & (STATE_CHECKED | STATE_SELECTED)) {
+      FireAccessibleFocusEvent(accessible, targetNode, aEvent);
     }
   }
   else if (eventType.LowerCaseEqualsLiteral("dommenubaractive"))
@@ -748,34 +809,21 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     FireCurrentFocusEvent();
   }
   else if (eventType.LowerCaseEqualsLiteral("domcontentloaded")) {
+    // Don't create the doc accessible until load scripts have a chance to set
+    // role attribute for <body> or <html> element, because the value of 
+    // role attribute will be cached when the doc accessible is Init()'d
     TryFireEarlyLoadEvent(accessible, targetNode);
   }
-  else {
-    // Menu popup events
-    PRUint32 menuEvent = 0;
-    if (eventType.LowerCaseEqualsLiteral("popupshown")) {
-      menuEvent = nsIAccessibleEvent::EVENT_MENUPOPUPSTART;
-    }
-    else if (eventType.EqualsLiteral("DOMMenuInactive")) {
-      menuEvent = nsIAccessibleEvent::EVENT_MENUPOPUPEND;
-    }
-    if (menuEvent) {
-      PRUint32 role = ROLE_NOTHING;
-      accessible->GetRole(&role);
-      if (role == ROLE_MENUPOPUP)
-        privAcc->FireToolkitEvent(menuEvent, accessible, nsnull);
+  else if (eventType.EqualsLiteral("DOMMenuInactive")) {
+    nsCOMPtr<nsIDOMXULPopupElement> popup(do_QueryInterface(targetNode));
+    if (popup) {
+      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_MENUPOPUPEND,
+                                accessible, nsnull);
     }
   }
 #else
   AtkStateChange stateData;
-  if (eventType.EqualsIgnoreCase("pageshow")) {
-    nsCOMPtr<nsIHTMLDocument> htmlDoc(do_QueryInterface(targetNode));
-    if (htmlDoc) {
-      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_REORDER, 
-                                accessible, nsnull);
-    }
-  }
-  else if (eventType.LowerCaseEqualsLiteral("focus") || 
+  if (eventType.LowerCaseEqualsLiteral("focus") || 
       eventType.LowerCaseEqualsLiteral("dommenuitemactive")) {
     if (treeItemAccessible) { // use focused treeitem
       privAcc = do_QueryInterface(treeItemAccessible);
@@ -785,13 +833,36 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     else if (anchorElement) {
       nsCOMPtr<nsIAccessibleHyperText> hyperText(do_QueryInterface(accessible));
       if (hyperText) {
+	nsCOMPtr<nsIDOMNode> focusedNode(do_QueryInterface(anchorElement));
+        NS_IF_RELEASE(gLastFocusedNode);
+        gLastFocusedNode = focusedNode;
+        NS_IF_ADDREF(gLastFocusedNode);
+
         PRInt32 selectedLink;
         hyperText->GetSelectedLinkIndex(&selectedLink);
         privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_ATK_LINK_SELECTED, accessible, &selectedLink);
       }
     }
+    else if (localName.EqualsIgnoreCase("radiogroup")) {
+      // fire focus event for checked radio instead of radiogroup
+      PRInt32 childCount = 0;
+      accessible->GetChildCount(&childCount);
+      nsCOMPtr<nsIAccessible> radioAcc;
+      for (PRInt32 index = 0; index < childCount; index++) {
+        accessible->GetChildAt(index, getter_AddRefs(radioAcc));
+        if (radioAcc) {
+          radioAcc->GetFinalState(&stateData.state);
+          if (stateData.state & (STATE_CHECKED | STATE_SELECTED)) {
+            break;
+          }
+        }
+      }
+      if (radioAcc) {
+        privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_FOCUS, radioAcc, nsnull);
+      }
+    }
     else
-      FireAccessibleFocusEvent(accessible, targetNode);
+      FireAccessibleFocusEvent(accessible, targetNode, aEvent);
   }
   else if (eventType.LowerCaseEqualsLiteral("select")) {
     if (treeItemAccessible) { // it's a XUL <tree>
@@ -800,6 +871,10 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
       privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_FOCUS, 
                                 treeItemAccessible, nsnull);
     }
+    else if (localName.LowerCaseEqualsLiteral("tabpanels")) {
+      // make GOK refresh "UI-Grab" window
+      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_REORDER, accessible, nsnull);
+    }
   }
 #if 0
   // XXX todo: value change events for ATK are done with 
@@ -816,11 +891,14 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
   else if (eventType.LowerCaseEqualsLiteral("checkboxstatechange") || // it's a XUL <checkbox>
            eventType.LowerCaseEqualsLiteral("radiostatechange")) { // it's a XUL <radio>
     accessible->GetFinalState(&stateData.state);
-    stateData.enable = (stateData.state & STATE_CHECKED) != 0;
+    // prefPane tab is implemented as list items in A11y, so we need to check STATE_SELECTED also
+    stateData.enable = (stateData.state & (STATE_CHECKED | STATE_SELECTED)) != 0;
     stateData.state = STATE_CHECKED;
     privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_STATE_CHANGE, accessible, &stateData);
-    if (eventType.LowerCaseEqualsLiteral("radiostatechange")) {
-      FireAccessibleFocusEvent(accessible, targetNode);
+    // only fire focus event for checked radio
+    if (eventType.LowerCaseEqualsLiteral("radiostatechange") &&
+        stateData.enable) {
+      FireAccessibleFocusEvent(accessible, targetNode, aEvent);
     }
   }
   else if (eventType.LowerCaseEqualsLiteral("openstatechange")) { // collapsed/expanded changed
@@ -846,10 +924,15 @@ NS_IMETHODIMP nsRootAccessible::HandleEvent(nsIDOMEvent* aEvent)
     FireCurrentFocusEvent();
   }
   else if (eventType.LowerCaseEqualsLiteral("popupshown")) {
-    FireAccessibleFocusEvent(accessible, targetNode);
+    FireAccessibleFocusEvent(accessible, targetNode, aEvent);
   }
   else if (eventType.EqualsLiteral("DOMMenuInactive")) {
-    //FireAccessibleFocusEvent(accessible, targetNode);  // Not yet used in ATK
+    nsCOMPtr<nsIDOMXULPopupElement> popup(do_QueryInterface(targetNode));
+    if (popup) {
+      privAcc->FireToolkitEvent(nsIAccessibleEvent::EVENT_MENUPOPUPEND,
+                                accessible, nsnull);
+    }
+
   }
 #endif
   return NS_OK;
@@ -971,3 +1054,61 @@ NS_IMETHODIMP nsRootAccessible::Shutdown()
   return nsDocAccessibleWrap::Shutdown();
 }
 
+already_AddRefed<nsIDocShellTreeItem>
+nsRootAccessible::GetContentDocShell(nsIDocShellTreeItem *aStart)
+{
+  PRInt32 itemType;
+  aStart->GetItemType(&itemType);
+  if (itemType == nsIDocShellTreeItem::typeContent) {
+    nsCOMPtr<nsIAccessibleDocument> accDoc = GetDocAccessibleFor(aStart);
+    nsCOMPtr<nsIAccessible> accessible = do_QueryInterface(accDoc);
+    // If ancestor chain of accessibles is not completely visible,
+    // don't use this one. This happens for example if it's inside
+    // a background tab (tabbed browsing)
+    while (accessible) {
+      PRUint32 state;
+      accessible->GetFinalState(&state);
+      if (state & STATE_INVISIBLE) {
+        return nsnull;
+      }
+      nsCOMPtr<nsIAccessible> ancestor;
+      accessible->GetParent(getter_AddRefs(ancestor));
+      accessible.swap(ancestor);
+    }
+
+    NS_ADDREF(aStart);
+    return aStart;
+  }
+  nsCOMPtr<nsIDocShellTreeNode> treeNode(do_QueryInterface(aStart));
+  if (treeNode) {
+    PRInt32 subDocuments;
+    treeNode->GetChildCount(&subDocuments);
+    for (PRInt32 count = 0; count < subDocuments; count ++) {
+      nsCOMPtr<nsIDocShellTreeItem> treeItemChild, contentTreeItem;
+      treeNode->GetChildAt(count, getter_AddRefs(treeItemChild));
+      NS_ENSURE_TRUE(treeItemChild, nsnull);
+      contentTreeItem = GetContentDocShell(treeItemChild);
+      if (contentTreeItem) {
+        NS_ADDREF(aStart = contentTreeItem);
+        return aStart;
+      }
+    }
+  }
+  return nsnull;
+}
+
+NS_IMETHODIMP nsRootAccessible::GetAccessibleRelated(PRUint32 aRelationType,
+                                                     nsIAccessible **aRelated)
+{
+  *aRelated = nsnull;
+
+  if (!mDOMNode || aRelationType != RELATION_EMBEDS) {
+    return nsDocAccessibleWrap::GetAccessibleRelated(aRelationType, aRelated);
+  }
+
+  nsCOMPtr<nsIDocShellTreeItem> treeItem = GetDocShellTreeItemFor(mDOMNode);   
+  nsCOMPtr<nsIDocShellTreeItem> contentTreeItem = GetContentDocShell(treeItem);
+  nsCOMPtr<nsIAccessibleDocument> accDoc = GetDocAccessibleFor(contentTreeItem);
+  return accDoc->QueryInterface(NS_GET_IID(nsIAccessible), (void**)aRelated);
+}
+
