@@ -53,7 +53,8 @@
 #include "nsAutoPtr.h"
 #include "nsIEventQueue.h"
 #include "nsEventQueueUtils.h"
-
+#include "nsIPrincipal.h"
+#include "nsIScriptSecurityManager.h"
 
 static PRUint32 gEntryID = 0;
 
@@ -96,6 +97,7 @@ nsSHEntry::nsSHEntry(const nsSHEntry &other)
   , mCacheKey(other.mCacheKey)
   , mParent(other.mParent)
   , mViewerBounds(0, 0, 0, 0)
+  , mOwner(other.mOwner)
 {
 }
 
@@ -123,7 +125,10 @@ nsSHEntry::~nsSHEntry()
 //    nsSHEntry: nsISupports
 //*****************************************************************************
 
-NS_IMPL_ISUPPORTS4(nsSHEntry, nsISHContainer, nsISHEntry, nsIHistoryEntry,
+NS_IMPL_ISUPPORTS6(nsSHEntry, nsISHContainer, nsISHEntry,
+                   nsISHEntry_MOZILLA_1_8_BRANCH,
+                   nsISHEntry_MOZILLA_1_8_BRANCH2,
+                   nsIHistoryEntry,
                    nsIDocumentObserver)
 
 //*****************************************************************************
@@ -400,12 +405,26 @@ nsSHEntry::Create(nsIURI * aURI, const nsAString &aTitle,
                   nsIInputStream * aInputStream,
                   nsILayoutHistoryState * aLayoutHistoryState,
                   nsISupports * aCacheKey, const nsACString& aContentType)
+{
+  return Create_MOZILLA_1_8_BRANCH(aURI, aTitle, aInputStream,
+                                   aLayoutHistoryState, aCacheKey,
+                                   aContentType, nsnull);
+}
+
+NS_IMETHODIMP
+nsSHEntry::Create_MOZILLA_1_8_BRANCH(nsIURI * aURI, const nsAString &aTitle,
+                                     nsIInputStream * aInputStream,
+                                     nsILayoutHistoryState * aLayoutHistoryState,
+                                     nsISupports * aCacheKey,
+                                     const nsACString& aContentType,
+                                     nsISupports* aOwner)
 {
   mURI = aURI;
   mTitle = aTitle;
   mPostData = aInputStream;
   mCacheKey = aCacheKey;
   mContentType = aContentType;
+  mOwner = aOwner;
     
   // Set the LoadType by default to loadHistory during creation
   mLoadType = (PRUint32) nsIDocShellLoadInfo::loadHistory;
@@ -484,6 +503,47 @@ nsSHEntry::GetViewerBounds(nsRect &aBounds)
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsSHEntry::GetOwner(nsISupports **aOwner)
+{
+  NS_IF_ADDREF(*aOwner = mOwner);
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsSHEntry::GetOwnerURI(nsIURI **aOwnerURI)
+{
+  nsCOMPtr<nsIPrincipal> prin(do_QueryInterface(mOwner));
+  if (prin) {
+    // Make sure not to hand out a ref to the principal's URI
+    nsCOMPtr<nsIURI> uri;
+    prin->GetURI(getter_AddRefs(uri));
+    if (uri) {
+      return uri->Clone(aOwnerURI);
+    }
+  }
+
+  *aOwnerURI = nsnull;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsSHEntry::SetOwnerURI(nsIURI *aOwnerURI)
+{
+  nsCOMPtr<nsIScriptSecurityManager> secMan =
+    do_GetService(NS_SCRIPTSECURITYMANAGER_CONTRACTID);
+  if (secMan && aOwnerURI) {
+    nsCOMPtr<nsIPrincipal> prin;
+    nsresult rv =
+      secMan->GetCodebasePrincipal(aOwnerURI, getter_AddRefs(prin));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    mOwner = prin;
+  }
+
+  return NS_OK;
+}
+
 //*****************************************************************************
 //    nsSHEntry: nsISHContainer
 //*****************************************************************************
